{
  GeoGet - knihovna GeoLocation

  Hledn omezujcch souadnic k zadanmu bodu na kouli a kruhu kolem nj.
  Dle http://janmatuschek.de/LatitudeLongitudeBoundingCoordinates
  Jan Philip Matuschek

  mpistora

  2011-02-04
  2011-02-06  0.5.0  prvn pouit
}

unit GeoLocation;

const
  PI = 3.1415926535897932385;
  TO_RADIANS = PI/180;
	MIN_LAT = -PI/2; // -90
	MAX_LAT =  PI/2; //  90
	MIN_LON = -PI;   // -180
	MAX_LON =  PI;   //  180

procedure BoundingCoordinates(Lat, Lon: extended; // bod v souadnicch ve stupnch
                              distance, radius: extended; // max. vzdlenost od bodu, polomr koule
                              var  minLat, minLon, maxLat, maxLon: extended); // vsledn ohraniujc souadnice
var
  radLat, radLon, radDist, deltaLon: extended;
begin
	if (radius < 0) or (distance < 0) then RaiseException(ErCustomError, 'Illegal Argument');

  radLat := Lat * TO_RADIANS;
	radLon := Lon * TO_RADIANS;
	if (radLat < MIN_LAT) or (radLat > MAX_LAT) or
		  (radLon < MIN_LON) or (radLon > MAX_LON) then RaiseException(ErCustomError, 'Illegal Argument');

	// angular distance in radians on a great circle
	radDist := distance / radius;

	minLat := radLat - radDist;
	maxLat := radLat + radDist;

	if (minLat > MIN_LAT) and (maxLat < MAX_LAT) then
	begin
		deltaLon := arcsin(sin(radDist) / cos(radLat));
		minLon := radLon - deltaLon;
		if (minLon < MIN_LON) then minLon := minLon + 2*PI;
		maxLon := radLon + deltaLon;
		if (maxLon > MAX_LON) then maxLon := maxLon - 2*PI;
  end
	else
	begin
		// a pole is within the distance
		if minLat < MIN_LAT then minLat := MIN_LAT;
		if maxLat > MAX_LAT then maxLat := MAX_LAT;
		minLon := MIN_LON;
		maxLon := MAX_LON;
  end;
  minLat :=  minLat / TO_RADIANS;
  minLon :=  minLon / TO_RADIANS;
  maxLat :=  maxLat / TO_RADIANS;
  maxLon :=  maxLon / TO_RADIANS;
end;

end.
