{
  GeoGet - knihovna TimeConversions

  Hledn asov zony v danm bod,
  pevod asu mezi znami,
  pevod do a z nkolika textovch formt.

  mpistora

  2011-01-09
  2011-01-25  0.5.0 prvn beta
  2011-02-01  0.5.1 oeten chyb
  2011-02-03  0.5.2 funkce TCUTCDateTimeToISOTime
  2011-02-06  0.5.3 monost nastavit znu PC (provoz bez webovch slueb)
  2011-02-15  1.0.0 prvn verze
}

unit TimeConversions;

const
  TC_ERR_DELIM = ' TCERR: '; // oddlova v textu vlastnch vjimek
  ETC_UTC = 'Etc/UTC'; // timezoneid pro UTC
var
  TCGeoNamesUserName: string;
  TCTimeZone: string;

// Nastaven konfigurace
procedure TCConfig(geonamesusername, timezone: string);
begin
  TCGeoNamesUserName := geonamesusername;
  TCTimeZone := timezone;
end;

// Vyvoln chyby
procedure TCError(level: byte; msg: string);
begin
  RaiseException(erCustomError, IntToStr(level) + TC_ERR_DELIM + msg); // kd oodlova text
end;

// Dekdovn kdu a textu chyby
procedure TCParseError(var level: byte; var msg: string);
var
  levstr: string;
begin
  if ExceptionType = erCustomError then
  begin
    msg := ExceptionParam;
    level := 0;
    if Pos(TC_ERR_DELIM, msg) > 0 then
    begin
      levstr := Fetch(msg,TC_ERR_DELIM);
      level := StrToInt(levstr);
    end;
  end
  else
    RaiseLastException;
end;

// Pevod datetime v ISO formtu (nap. 2010-12-31T18:34:55Z) na TDateTime a znak zny
function TCISOTimeToDateTimeZone(isotime: string; var zone: string): TDateTime;
var
  Year, Month, Day, Hour, Min, Sec, MSec: Word;
  se: extended;
begin
  Result := 0;
  if isotime = '' then Exit;
  Year := StrToInt(Fetch(isotime, '-'));
  Month := StrToInt(Fetch(isotime, '-'));
  Day := StrToInt(Fetch(isotime, 'T'));
  Hour := StrToInt(Fetch(isotime, ':'));
  Min := StrToInt(Fetch(isotime, ':'));
  zone := Copy(isotime, Length(isotime), 1);
  se := StrToFl(Fetch(isotime, zone));
  Sec := Trunc(se);
  MSec := Trunc(1000*(se-Sec));
  Result := EncodeDate(Year, Month, Day) + EncodeTime(Hour, Min, Sec, MSec);
end;

// Pevod UTC datetime v ISO formtu (nap. 2010-12-31T18:34:55Z) na TDateTime
function TCUTCISOTimeToDateTime(isotime: string): TDateTime;
var
  zone: string;
begin
  Result := TCISOTimeToDateTimeZone(isotime, zone);
end;

// Pevod UTC TDateTime na datetime v ISO formtu (nap. 2010-12-31T18:34:55Z)
function TCUTCDateTimeToISOTime(datetime: TDateTime): string;
begin
  Result := FormatDateTime('yyyy-mm-dd"T"hh:nn:ss"Z"', datetime);
end;

// Pevod msce v textovm formtu (nap. January) na slo 1 - 12
function TCENMonthToInt(entextmonth: string): integer;
begin
  case entextmonth of
    'January':   Result := 1;
    'February':  Result := 2;
    'March':     Result := 3;
    'April':     Result := 4;
    'May':       Result := 5;
    'June':      Result := 6;
    'July':      Result := 7;
    'August':    Result := 8;
    'September': Result := 9;
    'October':   Result := 10;
    'November':  Result := 11;
    'December':  Result := 12;
  end;
end;

// Pevod datetime v textovm formtu (nap. January 8, 2011) na TDateTime
function TCENTextTimeToDateTime(entexttime: string): TDateTime;
var
  year, month, day: string;
begin
  month := Fetch(entexttime, ' ');
  day := Fetch(entexttime, ', ');
  year := entexttime;
  Result := EncodeDate(StrToInt(year), TCENMonthToInt(month), StrToInt(day));
end;

// Pevod datetime s asovou znou na etzec (nap. 2010-04-24 16:39:42 Europe/Prague DST)
function TCDateTimeZoneToStr(datetime: TDateTime; timezone: string; dst: boolean): string;
begin
  Result := FormatDateTime('yyyy-mm-dd hh:nn:ss', datetime)
  if timezone <> '' then Result := Result + ' ' + timezone;
  if dst then Result := Result + ' DST';
end;

// Zjitn asov zny v danm bod
function TCPointToTimeZone(lat, lon: extended): string;
var
  host, par, user, resp: string;
begin
  if TCTimeZone <> '' then  // je-li v konfiguraci konstanta
  begin
    Result := TCTimeZone;   // tak ji stle vracej
    Exit;
  end;
  if TCGeoNamesUserName = '' then // nen-li v konfiguraci username na geonames.org
  begin
    host := 'ws'; // tak pouvej pvodn petenou free slubu
    user := '';
  end
  else // je-li v konfiguraci username na geonames.org
  begin
    host := 'api'; // tak pouvej novou slubu
    user := '&username=' + TCGeoNamesUserName; // ke kter se mus poslat username
  end;
  host := host + '.geonames.org';
  par := 'lat=' + FloatToStr(lat) + '&lng=' + FloatToStr(lon);
  if HttpAsk('GET', 'http://' + host + '/timezone?' + par + user, '', '', resp) then
  begin
    Fetch(resp, '<timezoneId>');
    Result := Fetch(resp, '</timezoneId>');
    if Result = '' then TCError(2, host + ' peten');
  end
  else TCError(2, host + ' neodpovd');
end;

// Pevod datetime z jedn asov zny do jin
procedure TCConvertTimeZone(datetime1: TDateTime; timezone1: string; var datetime2: TDateTime; timezone2: string; var dst: boolean);
var
  host, par, resp: string;
  er, ds, h, m, s, dat: string;
begin
  if (timezone1 = ETC_UTC) and (timezone2 = 'PC') then
  begin
    datetime2 := UTCtoLocal(datetime1);
    Exit;
  end;
  if (timezone1 = 'PC') and (timezone2 = ETC_UTC) then
  begin
    datetime2 := LocalToUTC(datetime1);
    Exit;
  end;
  host := 'www.timezoneconverter.com';
  par := 'use_current_datetime=0&month=' +
    FormatDateTime('m', datetime1) + '&day=' + FormatDateTime('d', datetime1) + '&year=' + FormatDateTime('yyyy', datetime1) +
    '&time=' + EncodeURLelement(FormatDateTime('hh:mm:ss', datetime1)) + '&time_type=24hour&fromzone=' + EncodeURLelement(timezone1) +
    '&tozone=' + EncodeURLelement(timezone2) + '&Submit=Convert';
  if HttpAsk('POST', 'http://' + host + '/cgi-bin/tzc.tzc', 'application/x-www-form-urlencoded',
      par, resp) then
  begin // parsuj strnku
    If Pos('<b>ERROR:</b>', resp) > 0 then
    begin
      Fetch(resp, '<b>ERROR:</b>');
      er := Fetch(resp, '<br>');
      TCError(2, Trim(er));
    end;
    Fetch(resp, 'To time:');
    Fetch(resp, 'DST ');
    ds := Fetch(resp, '"');
    Fetch(resp, 'converts to<br>');
    Fetch(resp, '<b>');
    h := Fetch(resp, ':');
    m := Fetch(resp, ':');
    s := Fetch(resp, ' ');
    Fetch(resp, ' ');
    dat := Fetch(resp, '</b>');
    datetime2 := TCENTextTimeToDateTime(dat) + EncodeTime(StrToInt(h), StrToInt(m), StrToInt(s), 0);
    dst := (ds = 'ON');
    if Abs(datetime2 - datetime1) > 2 then
      TCError(2, host + ' nevrtil pouitelnou odpovd');
  end
  else TCError(2, host + ' neodpovd');
end;

// Pevod UTC datetime do mstn asov zny bodu
procedure TCPointLocalTime(lat, lon: extended; utcdatetime: TDateTime; var pointlocaldatetime: TDateTime; var pointtimezone: string; var dst: boolean);
begin
  pointtimezone := TCPointToTimeZone(lat, lon);
  if pointtimezone <> '' then
    TCConvertTimeZone(utcdatetime, ETC_UTC, pointlocaldatetime, pointtimezone, dst);
end;

// Pevod mstnho datumu v danm bod na interval 00:00 - 24:00 v UTC datetime
procedure TCPointLocalDateToUTCDateTimeInterval(lat, lon: extended; pointlocaldate: TDateTime; var utcdatetime00: TDateTime; var utcdatetime24: TDateTime);
var
  pointtimezone: string;
  dst: boolean;
begin
  pointtimezone := TCPointToTimeZone(lat, lon);
  if pointtimezone <> '' then
    TCConvertTimeZone(Int(pointlocaldate), pointtimezone, utcdatetime00, ETC_UTC, dst);
    TCConvertTimeZone(Int(pointlocaldate) + 1, pointtimezone, utcdatetime24, ETC_UTC, dst);
end;

end.
