{
  GeoGet - knihovna TrackLogParser

  V danm GPX souboru s tracklogem najde nejbli zznam podle nkolika monch kritri.

  mpistora

  2011-01-14
  2011-01-25  0.5.0 prvn beta
  2011-02-01  0.5.1 oeten chyb
  2011-02-04  0.5.2 filtrovn bod dle asu pomoc XPath
  2011-02-06  0.5.3 filtrovn bod dle ohraniujcho boxu pomoc XPath
  2011-02-15  1.0.0 prvn verze
}

unit TrackLogParser;

uses
  TimeConversions, GeoLocation;
  
const
  MAX_DIST = 100000;
  
var
  Document: Variant;

function TLPSeparateRightBack(const Value, Delimiter: string): string;
begin
  Result := Reverse(SeparateLeft(Reverse(Value), Reverse(Delimiter)));
end;

function TLPSelectSingleNode(node: Variant; xpath: string): string;
begin
  try
    Result := node.SelectSingleNode(xpath).text;
  except
    Result := '';
  end;
end;

// V danm GPX souboru s tracklogem najde nejbli zznam podle nkolika monch kritri:
//  nejbli k zadanmu mstu
//   k tomu je jet mono doplnit maximln tolerovanou vzdlenost a/nebo asov interval
//  nebo nejbli k zadanmu asu
//   k tomu je jet mono doplnit maximln tolerovanou asovou odchylku
// Vrt True pokud najde poadovan zznam a o nm ve vstupnch argumentech:
//  souadnice, as a vku (pokud je zaznamenna) a dle
//  vzdlenost od zadanho bodu, resp. asu, ta se vrac i pokud se nenajde vyhovujc (pak je to tak minimln nalezen)
function TLPSearchNearstPoint(
  filepath, filename: string; // cesta a nzev GPX souboru
  lat, lon: extended; // souadnice zadanho msta
  requiredproximity: extended; // poadovan blzkost v km
  utcdatetimefrom: TDateTime; // zadan as v UTC, pp. zatek poadovanho intervalu
  utcdatetimeto: TDateTime; // konec poadovanho intervalu v UTC
  requireddifference: Double; // poadovan asov odchylka ve dnech
  var flat, flon: extended; // souadnice nalezenho bodu
  var utcdatetime: TDateTime; // UTC as nalezenho bodu
  var elevation: extended; // nadmosk vka nalezenho bodu v m
  var mindistance: extended; // vzdlenost nejlepho nalezenho bodu v km
  var mindifference: Double // asov odchylka nejlepho nalezenho bodu ve dnech
): boolean;
var
  filepn: string;
  NodeList: Variant;
  Node: Variant;
  n: integer;
  dist: extended;
  plat, plon: extended;
  isotime: string;
  ptime: TDateTime;
  diff: TDateTime;
  bypoint, interval, minp: boolean;
  namespaceURI, xpath, and180or: string;
  minLat, minLon, maxLat, maxLon: extended;
begin
  if VarIsEmpty(Document) then  // pi prvnm prchodu otevi OLE object
  begin
    try
      Document := CreateOleObject('Msxml2.DOMDocument.6.0');
    except
      Document := CreateOleObject('Microsoft.XMLDOM');
    end;
    Document.async := False;
    Document.validateOnParse := False;
    Document.resolveExternals := False;
    Document.preserveWhiteSpace := True;
    Document.setProperty('SelectionLanguage', 'XPath');
  end;
  if (lat <> 0) and (lon <> 0) then
  begin
    bypoint := True;
    interval := (utcdatetimefrom <> 0) and (utcdatetimeto <> 0);
    mindistance := MAX_DIST;
  end
  else
  begin
    if (utcdatetimefrom <> 0) and (utcdatetimeto = 0) then
    begin
      bypoint := False;
      mindifference := MAX_DIST;
    end
    else
      TCError(1, 'chybn parametry');
  end;
  if filepath = '' then
    filepn := filename
  else
    filepn := filepath + '\' + filename;
  if not FileExists(filepn) then
    TCError(1, 'soubor ' + filename + ' nenalezen');
  // nati GPX, resp. XML dokument
  Document.load(filepn);
  if Document.parseError.errorCode <> 0 then
    TCError(1, 'chyba ' + Document.parseError.reason + ' pri zpracovn souboru ' + filename);
  namespaceURI := Document.documentElement.namespaceURI;
  if not RegexFind('http://www.topografix.com/GPX/1/[01]', namespaceURI) then // nen to GPX 1.0 nebo 1.1
    TCError(1, 'souboru ' + filename + ' m nepodporovan default namespace ' + namespaceURI);
  Document.setProperty('SelectionNamespaces', 'xmlns:ns=''' + namespaceURI + '''');
  // sestav XPath dotaz
  xpath := '/ns:gpx/ns:trk/ns:trkseg/ns:trkpt'
  if bypoint and interval then  // interval: pi hledn v konkrtnm souboru neomezuj souadnice, aby se vypsal nejble nalezen bod
  begin
    BoundingCoordinates(lat, lon, // bod v souadnicch ve stupnch
                        requiredproximity, 6371.01, // max. vzdlenost od bodu, polomr koule
                        minLat, minLon, maxLat, maxLon);
    if minLon > maxLon then and180or := 'or' else and180or := 'and';
    xpath := xpath + '[(@lat>=' + FlToStr(minLat) + ' and @lat<='  + FlToStr(maxLat) + ') and (@lon >= ' +
                       + FlToStr(minLon) + ' ' + and180or + ' @lon<='  + FlToStr(maxLon) + ')]';
  end;
  if interval then // je-li zadn interval, vyber jen body s asem z intervalu
    xpath := xpath + '[translate(ns:time,"-T:Z","")>=' + FormatDateTime('yyyymmddhhnnss', utcdatetimefrom) +
             ' and translate(ns:time,"-T:Z","")<=' + FormatDateTime('yyyymmddhhnnss', utcdatetimeto) + ']'
  else  // nen-li, jen testuj existenci time
    xpath := xpath + '[ns:time]';
  NodeList := Document.DocumentElement.SelectNodes(xpath); // <--- XPath select
  // cyklus pes vybran body
  for n := 0 to NodeList.Length - 1 do
  begin
    Node := NodeList.item[n];
    minp := False;
    if bypoint then // hledn dle zadanhou msta
    begin
      plat := StrToFl(Node.getAttribute('lat'));
      plon := StrToFl(Node.getAttribute('lon'));
      dist := PointDistance(lat, lon, plat, plon);
      if dist < mindistance then
      begin
        mindistance := dist;
        if (requiredproximity = 0) or (dist <= requiredproximity) then
        begin
          minp := True;
          isotime := TLPSelectSingleNode(Node, 'ns:time');
          ptime := TCUTCISOTimeToDateTime(isotime);
        end
      end
    end
    else  // hledn dle zadanho asu
    begin
      isotime := TLPSelectSingleNode(Node, 'ns:time');
      ptime := TCUTCISOTimeToDateTime(isotime);
      diff := Abs(utcdatetimefrom - ptime);
      if diff < mindifference then
      begin
        if (requireddifference = 0) or (diff <= requireddifference) then
        begin
          minp := True;
          plat := StrToFl(Node.getAttribute('lat'));
          plon := StrToFl(Node.getAttribute('lon'));
        end
      end
    end;
    if minp then
    begin
      flat := plat;
      flon := plon;
      utcdatetime := ptime;
      elevation := StrToFl(TLPSelectSingleNode(Node, 'ns:ele'));
      Result := True;
    end;
    Node := unassigned; // bez tohoto neustle narst alokace pamti
  end;
  NodeList := unassigned;
end;

// V danm GPX souboru s tracklogem najde nejbli zznam
//  nejbli k zadanmu mstu
//   v maximln tolerovano vzdlenosti
// Vrt True pokud najde poadovan zznam a o nm ve vstupnch argumentech:
//  souadnice a as
//  vzdlenost od zadanho bodu
procedure TLPSearchNearstPointFile(
  filepath, filename: string; // cesta a nzev GPX souboru
  lat, lon: extended; // souadnice zadanho msta
  requiredproximity: extended; // poadovan blzkost v km
  var utcdatetime: TDateTime; // UTC as nalezenho bodu
  var mindistance: extended // vzdlenost nejlepho nalezenho bodu v km
);
var
  flat, flon: extended;
  elevation: extended;
  mindifference: Double;
begin
  if not TLPSearchNearstPoint(filepath, filename, lat, lon, requiredproximity,
      0, 0, 0, flat, flon, utcdatetime, elevation, mindistance, mindifference) then
    if mindistance = MAX_DIST then
      TCError(1, 'v souboru ' + filename + ' nebyl nalezen vyhovujc bod')
    else
      TCError(1, 'v souboru ' + filename + ' je nejbli bod ' + FormatFloat('0.0', mindistance*1000) + ' m' );
end;

// V danm adresi s GPX soubory s tracklogy najde nejbli zznam
//  nejbli k zadanmu mstu
//   v maximln tolerovano vzdlenosti
// Vrt True pokud najde poadovan zznam a o nm ve vstupnch argumentech:
//  souadnice a as
//  vzdlenost od zadanho bodu
procedure TLPSearchNearstPointDir(
  filepath: string; // adres s GPX soubory
  lat, lon: extended; // souadnice zadanho msta
  requiredproximity: extended; // poadovan blzkost v km
  utcdatetimefrom: TDateTime; // zatek poadovanho intervalu v UTC
  utcdatetimeto: TDateTime; // konec poadovanho intervalu v UTC
  var utcdatetime: TDateTime; // UTC as nalezenho bodu
  var mindistance: extended // vzdlenost nejlepho nalezenho bodu v km
);
var
  fullpath, filename: string;
  flat, flon: extended;
  fileutcdatetime: TDateTime;
  filemindistance: extended;
  elevation: extended;
  mindifference: Double;
  files: TStringList;
  n: integer;
  fnd: boolean;
begin
  if not DirectoryExists(filepath) then
    TCError(1, 'adres ' + filepath + ' nenalezen');
  files := TStringList.Create
  mindistance := MAX_DIST;
  FileList(filepath, files);
  fnd := False;
  for n := 0 to files.count - 1 do
  begin
    fullpath := files[n];
    if Lowercase(TLPSeparateRightBack(fullpath, '.')) = 'gpx' then
    begin
      filename := TLPSeparateRightBack(fullpath, '\');
      //GeoBusyKind(filename);
      if TLPSearchNearstPoint(filepath, filename, lat, lon, requiredproximity,
          utcdatetimefrom, utcdatetimeto, 0, flat, flon, fileutcdatetime, elevation, filemindistance, mindifference) then
      begin
        if filemindistance < mindistance then
        begin
          mindistance := filemindistance;
          utcdatetime := fileutcdatetime;
          fnd := True;
        end
      end
    end
  end;
  if not fnd then
    if mindistance = MAX_DIST then
      TCError(1, 'v adresi ' + TLPSeparateRightBack(filepath, '\') + ' nebyl nalezen vyhovujc bod')
    else
      TCError(1, 'v adresi ' + TLPSeparateRightBack(filepath, '\') + ' je nejbli bod ' + FormatFloat('0.0', mindistance*1000) + ' m' );
end;

end.
